@@ -1,4 +1,4 @@
-/*	$OpenBSD: rde_update.c,v 1.30 2004/08/20 15:31:56 claudio Exp $ */
+/*	$OpenBSD: rde_update.c,v 1.35 2005/03/11 12:54:20 claudio Exp $ */
 
 /*
  * Copyright (c) 2004 Claudio Jeker <claudio@openbsd.org>
@@ -278,7 +278,7 @@ up_generate_updates(struct rde_peer *peer,
 			/*
 			 * route reflector redistribution rules:
 			 * 1. if announce is set		-> announce
-			 * 2. old non-client, new non-client	-> no 
+			 * 2. old non-client, new non-client	-> no
 			 * 3. old client, new non-client	-> yes
 			 * 4. old non-client, new client	-> yes
 			 * 5. old client, new client		-> yes
@@ -320,7 +320,7 @@ up_generate_updates(struct rde_peer *peer,
 			return;
 
 		/*
-		 * Don't send messages back to originator 
+		 * Don't send messages back to originator
 		 * this is not specified in the RFC but seems logical.
 		 */
 		if ((attr = attr_optget(old->aspath,
@@ -334,6 +334,8 @@ up_generate_updates(struct rde_peer *peer,
 		/* copy attributes for output filter */
 		fasp = path_copy(old->aspath);
 
+		/* default override not needed here as this is a withdraw */
+
 		pt_getaddr(old->prefix, &addr);
 		if (rde_filter(peer, fasp, &addr,
 		    old->prefix->prefixlen, DIR_OUT) == ACTION_DENY) {
@@ -373,7 +375,7 @@ up_generate_updates(struct rde_peer *peer,
 			/*
 			 * route reflector redistribution rules:
 			 * 1. if announce is set		-> announce
-			 * 2. old non-client, new non-client	-> no 
+			 * 2. old non-client, new non-client	-> no
 			 * 3. old client, new non-client	-> yes
 			 * 4. old non-client, new client	-> yes
 			 * 5. old client, new client		-> yes
@@ -431,6 +433,13 @@ up_generate_updates(struct rde_peer *peer,
 		/* copy attributes for output filter */
 		fasp = path_copy(new->aspath);
 
+		/*
+		 * apply default outgoing overrides,
+		 * actually only prepend-self
+		 */
+		rde_apply_set(fasp, &peer->conf.attrset, new->prefix->af,
+		    fasp->peer, DIR_DEFAULT_OUT);
+
 		pt_getaddr(new->prefix, &addr);
 		if (rde_filter(peer, fasp, &addr,
 		    new->prefix->prefixlen, DIR_OUT) == ACTION_DENY) {
@@ -440,7 +449,7 @@ up_generate_updates(struct rde_peer *peer,
 		}
 
 		/*
-		 * Don't send messages back to originator 
+		 * Don't send messages back to originator
 		 * this is not specified in the RFC but seems logical.
 		 */
 		if ((attr = attr_optget(new->aspath,
@@ -571,7 +580,12 @@ up_generate_attr(struct rde_peer *peer, struct update_attr *upa,
 	wlen += r; len -= r;
 
 	/* aspath */
-	path = aspath_prepend(a->aspath, rde_local_as(), peer->conf.ebgp);
+	if (!peer->conf.ebgp ||
+	    rde_decisionflags() & BGPD_FLAG_DECISION_TRANS_AS)
+		path = aspath_prepend(a->aspath, rde_local_as(), 0);
+	else
+		path = aspath_prepend(a->aspath, rde_local_as(), 1);
+
 	if ((r = attr_write(up_attr_buf + wlen, len, ATTR_WELL_KNOWN,
 	    ATTR_ASPATH, path->data, path->len)) == -1)
 		return (-1);
@@ -579,13 +593,19 @@ up_generate_attr(struct rde_peer *peer, struct update_attr *upa,
 	wlen += r; len -= r;
 
 	/* nexthop, already network byte order */
-	if (peer->conf.ebgp == 0) {
+	if (a->flags & F_NEXTHOP_NOMODIFY) {
+		/* no modify flag set */
+		if (a->nexthop == NULL)
+			nexthop = peer->local_v4_addr.v4.s_addr;
+		else
+			nexthop = a->nexthop->exit_nexthop.v4.s_addr;
+	} else if (!peer->conf.ebgp) {
 		/*
-		 * If directly connected use peer->local_addr
+		 * If directly connected use peer->local_v4_addr
 		 * this is only true for announced networks.
 		 */
 		if (a->nexthop == NULL)
-			nexthop = peer->local_addr.v4.s_addr;
+			nexthop = peer->local_v4_addr.v4.s_addr;
 		else if (a->nexthop->exit_nexthop.v4.s_addr ==
 		    peer->remote_addr.v4.s_addr)
 			/*
@@ -593,7 +613,7 @@ up_generate_attr(struct rde_peer *peer, struct update_attr *upa,
 			 * the nexthop set the nexthop to our local address.
 			 * This reduces the risk of routing loops.
 			 */
-			nexthop = peer->local_addr.v4.s_addr;
+			nexthop = peer->local_v4_addr.v4.s_addr;
 		else
 			nexthop = a->nexthop->exit_nexthop.v4.s_addr;
 	} else if (peer->conf.distance == 1) {
@@ -607,9 +627,9 @@ up_generate_attr(struct rde_peer *peer, struct update_attr *upa,
 				/* nexthop and peer are in the same net */
 				nexthop = a->nexthop->exit_nexthop.v4.s_addr;
 			else
-				nexthop = peer->local_addr.v4.s_addr;
+				nexthop = peer->local_v4_addr.v4.s_addr;
 		} else
-			nexthop = peer->local_addr.v4.s_addr;
+			nexthop = peer->local_v4_addr.v4.s_addr;
 	} else
 		/* ebgp multihop */
 		/*
@@ -617,7 +637,7 @@ up_generate_attr(struct rde_peer *peer, struct update_attr *upa,
 		 * NEXTHOP_CONNECTED set so it should be possible to unify the
 		 * two ebgp cases. But this is save and RFC compliant.
 		 */
-		nexthop = peer->local_addr.v4.s_addr;
+		nexthop = peer->local_v4_addr.v4.s_addr;
 
 	if ((r = attr_write(up_attr_buf + wlen, len, ATTR_WELL_KNOWN,
 	    ATTR_NEXTHOP, &nexthop, 4)) == -1)
